"""
scmdata utility functions
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Iterable

if TYPE_CHECKING:
    import scmdata.run


def scmrun_as_dict(
    inp: scmdata.run.BaseScmRun, groups: Iterable[str], separator: str = "_"
) -> Dict[str, scmdata.run.BaseScmRun]:
    """
    Group an :obj:`scmdata.run.BaseScmRun` into a dictionary with keys

    Parameters
    ----------
    inp
        :obj:`scmdata.run.BaseScmRun` to group

    groups
        Metadata keys to use to make the groups

    separator
        Separator for metadata values when making the keys

    Returns
    -------
        Grouped ``inp``
    """
    res = {}
    for inp_g in inp.groupby(groups):
        key = separator.join([inp_g.get_unique_meta(g, True) for g in groups])
        res[key] = inp_g

    return res



"""
Support for plotting during scipy optimisation
"""
from __future__ import annotations

import logging
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Protocol,
    Tuple,
    Union,
)

import numpy as np
import scmdata
from attrs import define, field

if TYPE_CHECKING:
    import attr
    import matplotlib
    import numpy.typing as nptype
    import pandas as pd
    import scmdata.run
    import tqdm

    import openscm_calibration.store


logger: logging.Logger = logging.getLogger(__name__)
"""Logger for this module"""


class SupportsScipyOptCallback(Protocol):
    """
    Class that supports being used as a callback with Scipy's optimisers
    """

    def callback_minimize(
        self,
        xk: nptype.NDArray[Union[np.float_, np.int_]],  # pylint:disable=invalid-name
    ) -> None:
        """
        Get cost of parameter vector

        This callback is intended to be used with `scipy.optimize.minimize`

        Parameters
        ----------
        xk
            Last used parameter vector
        """

    def callback_differential_evolution(
        self,
        xk: nptype.NDArray[Union[np.float_, np.int_]],  # pylint:disable=invalid-name
        convergence: Optional[float] = None,
    ) -> None:
        """
        Get cost of parameter vector

        This callback is intended to be used with
        `scipy.optimize.differential_evolution`

        Parameters
        ----------
        xk
            Parameter vector with best solution found so far

        convergence
            Received from :func:`scipy.optimize.differential_evolution`
            on callback. We are not sure what this does is or is used for.
        """


class SupportsFigUpdate(Protocol):  # pylint:disable=too-few-public-methods
    """
    Class that supports updating figures

    For example, :class:`IPython.core.display_functions.DisplayHandle`
    """

    def update(
        self,
        fig: matplotlib.figure.Figure,
    ) -> None:
        """
        Update the figure

        Parameters
        ----------
        fig
            Figure to update
        """


ScmRunToDictConverter = Callable[
    [scmdata.run.BaseScmRun], Dict[str, scmdata.run.BaseScmRun]
]
"""
Type hint helper for functions used as ``convert_scmrun_to_plot_dict``

Such functions convert an :obj:`scmdata.run.BaseScmRun` to a dictionary of
:obj:`scmdata.run.BaseScmRun`
"""


def _all_in_axes(
    instance: OptPlotter,  # pylint: disable=unused-argument
    attribute: attr.Attribute[Tuple[str]],
    value: Tuple[str],
) -> None:
    """
    Check all values are present in ``self.axes``

    Parameters
    ----------
    self
        Object instance

    attribute
        Attribute to check

    value
        Value to check

    Raises
    ------
    ValueError
        Not all elements in ``value`` are keys in ``self.axes``
    """
    missing_from_axes = [k for k in value if k not in instance.axes]
    if missing_from_axes:
        error_msg = (
            f"Error setting '{attribute.name}'. "
            "The following keys are missing from ``self.axes``: "
            f"``{missing_from_axes}``. "
            f"``self.axes.keys()`` is ``{instance.axes.keys()}``."
        )
        raise KeyError(error_msg)


def _compatible_with_convert_and_target(
    instance: OptPlotter,  # pylint: disable=unused-argument
    attribute: attr.Attribute[Tuple[str]],
    value: Tuple[str],
) -> None:
    """
    Check that the values are compatible with the target and the ScmRun conversion

    Specifically, compatible with ``self.target`` and
    ``self.convert_scmrun_to_plot_dict``

    Parameters
    ----------
    self
        Object instance

    attribute
        Attribute to check

    value
        Value to check

    Raises
    ------
    ValueError
        If the values in ``value`` are not all keys in the dictionary that is
        returned when ``self.convert_scmrun_to_plot_dict`` is applied to
        ``self.target``.
    """
    target_converted = instance.convert_scmrun_to_plot_dict(instance.target)
    missing_from_target_converted = [k for k in value if k not in target_converted]
    if missing_from_target_converted:
        error_msg = (
            f"Error setting '{attribute.name}'. "
            "The following keys are missing when "
            "``self.convert_scmrun_to_plot_dict`` is applied to "
            f"``self.target``: ``{missing_from_target_converted}``. "
            f"``target_converted.keys()`` is ``{target_converted.keys()}``."
        )
        raise ValueError(error_msg)


@define
class OptPlotter:
    """
    Optimisation progress plotting helper

    This class is an adapter between interfaces required by Scipy's callback
    arguments and updating the plots. The class holds all the information
    required to make useful plots. It is intended to be used in interactive
    Python i.e. to make updating plots.
    """

    holder: SupportsFigUpdate
    """Figure updater, typically :obj:`IPython.core.display_functions.DisplayHandle`"""

    fig: matplotlib.figure.Figure
    """Figure on which to plot"""

    axes: Dict[str, matplotlib.axes.Axes]
    """
    Dictionary storing axes on which to plot

    The plot of the cost function over time will be plotted on the axes with
    key given by ``cost_key``.

    Each parameter will be plotted on the axes with the same key as the
    parameter (as defined in ``parameters``)

    The timeseries will be plotted on the axes specified by
    ``timeseries_axes``. See docstring of ``timeseries_axes`` for rules about
    its values.
    """

    cost_key: str
    """Key for the axes on which the cost function should be plotted"""

    parameters: Tuple[str] = field(validator=[_all_in_axes])
    """
    Parameters to be optimised

    This must match the order in which the parameters are handled by the
    optimiser i.e. it is used to translate the unlabeled array of parameter
    values onto the desired axes.
    """

    timeseries_axes: Tuple[str] = field(
        validator=[_all_in_axes, _compatible_with_convert_and_target]
    )
    """
    Axes on which to plot timeseries

    The timeseries in ``target`` and ``store.res`` are
    converted into dictionaries using ``convert_scmrun_to_plot_dict``. The
    keys of the result of ``convert_scmrun_to_plot_dict`` must match the
    values in ``timeseries_axes``.
    """

    convert_scmrun_to_plot_dict: ScmRunToDictConverter
    """
    Callable which converts :obj:`scmdata.run.BaseScmRun` into a dictionary in
    which the keys are a subset of the values in ``timeseries_axes``
    """

    target: scmdata.run.BaseScmRun
    """Target to which we are optimising"""

    store: openscm_calibration.store.OptResStore
    """Optimisation result store"""

    thin_ts_to_plot: int = 20
    """
    Thinning to apply to the timeseries to plot

    In other words, only plot every ``thin_ts_to_plot`` runs on the timeseries
    plots. Plotting all runs can be very expensive.
    """

    plot_cost_kwargs: Optional[Dict[str, Any]] = None
    """Kwargs to pass to :func:`plot_costs`."""

    plot_parameters_kwargs: Optional[Dict[str, Any]] = None
    """Kwargs to pass to :func:`plot_parameters`."""

    plot_timeseries_kwargs: Optional[Dict[str, Any]] = None
    """Kwargs to pass to :func:`plot_timeseries`."""

    get_timeseries: Optional[Callable[[scmdata.run.BaseScmRun], pd.DataFrame]] = None
    """
    Function which converts :obj:`scmdata.run.BaseScmRun` into a :obj:`pd.DataFrame` for plotting

    If not provided, :func:`get_timeseries_default` is used
    """

    def callback_minimize(
        self,
        xk: nptype.NDArray[  # pylint:disable=invalid-name, unused-argument
            Union[np.float_, np.int_]
        ],
    ) -> None:
        """
        Update the plots

        Intended to be used as the `callback` argument to
        `scipy.optimize.minimize`

        Parameters
        ----------
        xk
            Last used parameter vector
        """
        self.update_plots()

    def callback_differential_evolution(
        self,
        xk: nptype.NDArray[  # pylint:disable=invalid-name, unused-argument
            Union[np.float_, np.int_]
        ],
        convergence: Optional[float] = None,  # pylint:disable=unused-argument
    ) -> None:
        """
        Update the plots

        Intended to be used as the `callback` argument to
        `scipy.optimize.differential_evolution`

        Parameters
        ----------
        xk
            Parameter vector with best solution found so far

        convergence
            Received from :func:`scipy.optimize.differential_evolution`
            on callback. We are not sure what this does is or what it is used
            for.
        """
        self.update_plots()

    def update_plots(self) -> None:
        """
        Update all the plots
        """
        costs, para_vals, res = self.store.get_costs_labelled_xsamples_res()

        # check if anything to plot
        if np.all(~np.isfinite(costs)):
            logger.info("No runs succeeded, nothing to plot")
            return

        # plot cost function
        ax_cost = self.axes[self.cost_key]
        ax_cost.clear()

        plot_cost_kwargs = (
            self.plot_cost_kwargs if self.plot_cost_kwargs is not None else {}
        )
        plot_costs(ax=ax_cost, ylabel=self.cost_key, costs=costs, **plot_cost_kwargs)

        # plot parameters
        for parameter in self.parameters:
            self.axes[parameter].clear()

        plot_parameters_kwargs = (
            self.plot_parameters_kwargs
            if self.plot_parameters_kwargs is not None
            else {}
        )
        plot_parameters(
            axes=self.axes,
            para_vals=para_vals,
            **plot_parameters_kwargs,
        )

        # plot timeseries
        best_run, others_to_plot = get_runs_to_plot(costs, res, self.thin_ts_to_plot)

        for k in self.timeseries_axes:
            self.axes[k].clear()

        plot_timeseries_kwargs = (
            self.plot_timeseries_kwargs
            if self.plot_timeseries_kwargs is not None
            else {}
        )

        if self.get_timeseries is not None:
            get_timeseries = self.get_timeseries
        else:
            get_timeseries = get_timeseries_default

        plot_timeseries(
            best_run=best_run,
            others_to_plot=others_to_plot,
            target=self.target,
            convert_scmrun_to_plot_dict=self.convert_scmrun_to_plot_dict,
            timeseries_keys=self.timeseries_axes,
            axes=self.axes,
            get_timeseries=get_timeseries,
            **plot_timeseries_kwargs,
        )

        # update and return
        self.fig.tight_layout()
        self.holder.update(self.fig)


def get_timeseries_default(
    inp: scmdata.run.BaseScmRun,
    time_axis: str = "year-month",
) -> pd.DataFrame:
    """
    Get timeseries for plotting from an :obj:`scmdata.run.BaseScmRun`

    This is the default function for doing this conversion in this module

    Parameters
    ----------
    inp
        :obj:`scmdata.run.BaseScmRun` to convert

    time_axis
        Passed to :meth:`inp.timeseries` when doing the
        conversion

    Returns
    -------
        Data with the time axis as rows, ready for simplified plotting using
        panda's plotting methods.
    """
    return inp.timeseries(time_axis=time_axis).T


def plot_costs(  # pylint:disable=too-many-arguments
    ax: matplotlib.axes.Axes,  # pylint:disable=invalid-name
    ylabel: str,
    costs: Tuple[float, ...],
    ymin: float = 0.0,
    get_ymax: Optional[Callable[[Tuple[float, ...]], float]] = None,
    alpha: float = 0.7,
    **kwargs: Any,
) -> None:
    r"""
    Plot cost function

    Parameters
    ----------
    ax
        Axes on which to plot

    ylabel
        y-axis label

    costs
        Costs to plot

    ymin
        Minimum y-axis value

    get_ymax
        Function which gets the y-max based on the costs. If not provided,
        :func:`get_ymax_default` is used

    alpha
        Alpha to apply to plotted points

    **kwargs
        Passed to :meth:`ax.scatter`
    """
    if get_ymax is None:
        get_ymax = get_ymax_default

    ax.scatter(range(len(costs)), costs, alpha=alpha, **kwargs)
    ax.set_ylabel(ylabel)

    ymax = get_ymax(costs)
    if not np.isfinite(ymax):
        ymax = 10**3

    ax.set_ylim(
        ymin=ymin,
        ymax=ymax,
    )


def get_ymax_default(
    costs: Tuple[float, ...],
    min_scale_factor: float = 10.0,
    min_v_median_scale_factor: float = 2.0,
) -> float:
    r"""
    Get y-max based on costs

    This is the default function used by :func:`plot_costs`. The algorithm is

    .. math::

        \text{ymax} = min(
            \text{min_scale_factor} \times min(costs),
            max(
                median(costs),
                \text{min_v_median_scale_factor} \times min(costs)
            )
        )

    Parameters
    ----------
    costs
        Cost function values

    min_scale_factor
        Value by which the minimum value is scaled when determining the plot
        limits

    min_v_median_scale_factor
        Value by which the minimum value is scaled when comparing to the
        median as part of determining the plot limits

    Returns
    -------
        Maximum value to use on the y-axis
    """
    min_cost = np.min(costs)
    ymax = np.min(
        [
            min_scale_factor * min_cost,
            np.max([np.median(costs), min_v_median_scale_factor * min_cost]),
        ]
    )

    return float(ymax)


def plot_parameters(
    axes: Dict[str, matplotlib.axes.Axes],
    para_vals: Dict[str, nptype.NDArray[Union[np.float_, np.int_]]],
    alpha: float = 0.7,
    **kwargs: Any,
) -> None:
    """
    Plot parameters

    Parameters
    ----------
    axes
        Axes on which to plot. The keys should match the keys in ``para_vals``

    para_vals
        Parameter values. Each key should be the name of a parameter

    alpha
        Alpha to use when calling :meth:`matplotlib.axes.Axes.scatter`

    **kwargs
        Passed to each call to :meth:`matplotlib.axes.Axes.scatter`
    """
    for parameter, values in para_vals.items():
        axes[parameter].scatter(range(len(values)), values, alpha=alpha, **kwargs)
        axes[parameter].set_ylabel(parameter)


DEFAULT_PLOT_TIMESERIES_BACKGROUND_TS_KWARGS: Dict[str, Any] = {
    "legend": False,
    "linewidth": 0.5,
    "alpha": 0.3,
    "color": "tab:gray",
    "zorder": 1.5,
}
"""
Default value for ``background_ts_kwargs`` used by ``plot_timeseries``

Provided to give the user an easy to modify these defaults if they wish or a
starting point
"""


DEFAULT_PLOT_TIMESERIES_TARGET_TS_KWARGS: Dict[str, Any] = {
    "legend": False,
    "linewidth": 2,
    "alpha": 0.8,
    "color": "tab:blue",
    "zorder": 2,
}
"""
Default value for ``target_ts_kwargs`` used by ``plot_timeseries``

Provided to give the user an easy to modify these defaults if they wish or a
starting point
"""


DEFAULT_PLOT_TIMESERIES_BEST_TS_KWARGS: Dict[str, Any] = {
    "legend": False,
    "linewidth": 2,
    "alpha": 0.8,
    "color": "tab:orange",
    "zorder": 2.5,
}
"""
Default value for ``best_ts_kwargs`` used by ``plot_timeseries``

Provided to give the user an easy to modify these defaults if they wish or a
starting point
"""


def plot_timeseries(  # pylint:disable=too-many-arguments,too-many-locals
    best_run: scmdata.run.BaseScmRun,
    others_to_plot: scmdata.run.BaseScmRun,
    target: scmdata.run.BaseScmRun,
    convert_scmrun_to_plot_dict: Callable[
        [scmdata.run.BaseScmRun], Dict[str, scmdata.run.BaseScmRun]
    ],
    timeseries_keys: Iterable[str],
    axes: Dict[str, matplotlib.axes.Axes],
    get_timeseries: Callable[[scmdata.run.BaseScmRun], pd.DataFrame],
    background_ts_kwargs: Optional[Dict[str, Any]] = None,
    target_ts_kwargs: Optional[Dict[str, Any]] = None,
    best_ts_kwargs: Optional[Dict[str, Any]] = None,
    ylabel_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """
    Plot timeseries

    This plots the target, the best result so far and other results

    Parameters
    ----------
    best_run
        Best run from iterations

    others_to_plot
        Other results to plot from iterations

    target
        Target timeseries

    convert_scmrun_to_plot_dict
        Callable which converts :obj:`scmdata.run.BaseScmRun` into a
        dictionary in which the keys are a subset of the values in
        ``axes``

    timeseries_keys
        Keys of the timeseries to plot

    axes
        Axes on which to plot

    get_timeseries
        Function which converts :obj:`scmdata.run.BaseScmRun` into a
        :obj:`pd.DataFrame` for plotting

    background_ts_kwargs
        Passed to :meth:`pd.DataFrame.plot.line` when plotting the background
        timeseries. If not supplied, we use
        ``DEFAULT_PLOT_TIMESERIES_BACKGROUND_TS_KWARGS``.

    target_ts_kwargs
        Passed to :meth:`pd.DataFrame.plot.line` when plotting the target
        timeseries. If not supplied, we use
        ``DEFAULT_PLOT_TIMESERIES_TARGET_TS_KWARGS``.

    best_ts_kwargs
        Passed to :meth:`pd.DataFrame.plot.line` when plotting the best
        timeseries. If not supplied, we use
        ``DEFAULT_PLOT_TIMESERIES_BEST_TS_KWARGS``.

    ylabel_kwargs
        Passed to :meth:`ax.set_ylabel` when setting the y-labels of each panel
    """
    if not background_ts_kwargs:
        background_ts_kwargs = DEFAULT_PLOT_TIMESERIES_BACKGROUND_TS_KWARGS

    if not target_ts_kwargs:
        target_ts_kwargs = DEFAULT_PLOT_TIMESERIES_TARGET_TS_KWARGS

    if not best_ts_kwargs:
        best_ts_kwargs = DEFAULT_PLOT_TIMESERIES_BEST_TS_KWARGS

    if not ylabel_kwargs:
        ylabel_kwargs = {}

    best_run_d = convert_scmrun_to_plot_dict(best_run)
    others_to_plot_d = convert_scmrun_to_plot_dict(others_to_plot)
    target_runs = convert_scmrun_to_plot_dict(target)

    for k in timeseries_keys:
        ax = axes[k]  # pylint: disable=invalid-name
        best_k = best_run_d[k]
        background_runs = others_to_plot_d[k]
        model_unit = background_runs.get_unique_meta("unit", True)

        target_k = target_runs[k]
        target_k_unit = target_k.get_unique_meta("unit", True)
        if target_k_unit != model_unit:
            # Avoidable user side, hence warn (see
            # https://docs.python.org/3/howto/logging.html#when-to-use-logging)
            warn_msg = (
                f"Converting target units ('{target_k_unit}') to model output "
                f"units ('{model_unit}'), this will happen every time you "
                "plot and is slow. Please convert the target units to the "
                "model's units before doing the optimisation for increased "
                "performance (the function "
                "``convert_target_to_model_output_units`` may be helpful)."
            )
            warnings.warn(warn_msg)
            target_k = target_k.convert_unit(model_unit)

        if not background_runs.empty:
            get_timeseries(background_runs).plot.line(
                ax=ax,
                **background_ts_kwargs,
            )

        get_timeseries(target_k).plot.line(
            ax=ax,
            **target_ts_kwargs,
        )

        get_timeseries(best_k).plot.line(
            ax=ax,
            **best_ts_kwargs,
        )

        ax.set_ylabel(k, **ylabel_kwargs)


def get_runs_to_plot(
    costs: Tuple[float, ...],
    res: Tuple[scmdata.run.BaseScmRun, ...],
    thin_ts_to_plot: int,
) -> Tuple[scmdata.run.BaseScmRun, scmdata.run.BaseScmRun]:
    """
    Get runs to plot

    This retrieves the run which best matches the target (has lowest cost) and
    then a series of others to plot

    Parameters
    ----------
    costs
        Cost function value for each run (used to determine the best result)

    res
        Results of each run. It is assumed that the elements in ``res`` and
        ``costs`` line up i.e. the nth element of ``costs`` is the cost
        function for the nth element in ``res``

    thin_ts_to_plot
        Thinning to apply to the timeseries to plot

        In other words, only plot every ``thin_ts_to_plot`` runs on the
        timeseries plots. Plotting all runs can be very expensive.

    Returns
    -------
        Best iteration and then other runs to plot

    Raises
    ------
    ValueError
        No successful runs are included in ``res``
    """
    # Convert to dict for quicker lookup later
    res_d_success = {i: v for i, v in enumerate(res) if v is not None}
    if not res_d_success:
        raise ValueError("No successful runs, please check")

    best_it = int(np.argmin(costs))
    out_best = res_d_success[best_it]

    success_keys = list(res_d_success.keys())
    to_plot_not_best = success_keys[
        len(success_keys) - 1 :: -thin_ts_to_plot  # noqa: E203 (prefer black)
    ]
    if best_it in to_plot_not_best:
        to_plot_not_best.remove(best_it)

    res_not_best = [res_d_success[i] for i in to_plot_not_best]
    out_not_best = scmdata.run_append(res_not_best)

    return out_best, out_not_best


@define
class CallbackProxy:
    """
    Callback helper

    This class acts as a proxy and decides whether the real callback should
    actually be called. If provided, it also keeps track of the number of
    model calls via a progress bar.
    """

    real_callback: SupportsScipyOptCallback
    """Callback to be called if a sufficient number of runs have been done"""

    store: openscm_calibration.store.OptResStore
    """Optimisation result store"""

    last_callback_val: int = 0
    """Number of model calls at last callback"""

    update_every: int = 50
    """Update the plots every X calls to the model"""

    progress_bar: Optional[tqdm.std.tqdm] = None
    """Progress bar to track iterations"""

    def callback_minimize(
        self,
        xk: nptype.NDArray[Union[np.float_, np.int_]],  # pylint:disable=invalid-name
    ) -> None:
        """
        Update the plots

        Intended to be used as the `callback` argument to
        `scipy.optimize.minimize`

        Parameters
        ----------
        xk
            Last used parameter vector
        """
        if self.time_to_call_real_callback():
            self.real_callback.callback_minimize(xk)

    def callback_differential_evolution(
        self,
        xk: nptype.NDArray[Union[np.float_, np.int_]],  # pylint:disable=invalid-name
        convergence: Optional[float] = None,
    ) -> None:
        """
        Update the plots

        Intended to be used as the `callback` argument to
        `scipy.optimize.differential_evolution`

        Parameters
        ----------
        xk
            Parameter vector with best solution found so far

        convergence
            Received from :func:`scipy.optimize.differential_evolution`
            on callback. Not sure what this does is or is used for.
        """
        if self.time_to_call_real_callback():
            self.real_callback.callback_differential_evolution(xk, convergence)

    def time_to_call_real_callback(self) -> bool:
        """
        Check whether it is time to call the real callback

        Returns
        -------
            ``True`` if the real callback should be called
        """
        n_model_calls = sum(x is not None for x in self.store.x_samples)
        if self.progress_bar:
            self.update_progress_bar(n_model_calls)

        if n_model_calls < self.last_callback_val + self.update_every:
            return False

        # Note that we ran the full callback
        self.last_callback_val = n_model_calls

        return True

    def update_progress_bar(self, n_model_calls: int) -> None:
        """
        Update the progress bar

        Parameters
        ----------
        n_model_calls
            Number of model calls in total

        Raises
        ------
        ValueError
            ``self.progress_bar`` is not set
        """
        if not self.progress_bar:
            raise ValueError("``self.progress_bar`` is not set")

        self.progress_bar.update(n_model_calls - self.progress_bar.last_print_n)


def convert_target_to_model_output_units(
    *,
    target: scmdata.run.BaseScmRun,
    model_output: scmdata.run.BaseScmRun,
    convert_scmrun_to_plot_dict: ScmRunToDictConverter,
) -> scmdata.run.BaseScmRun:
    """
    Convert the target data to the model output's units

    This is a helper function that allows the data to be lined up before
    setting up plotting etc.

    Parameters
    ----------
    target
        Target data to convert

    model_output
        A sample of the model's output

    convert_scmrun_to_plot_dict
        The function that will be used to convert
        :obj:`scmdata.run.BaseScmRun` to a dictionary when doing the plotting

    Returns
    -------
        Target data with units that match the model output
    """
    target_d = convert_scmrun_to_plot_dict(target)
    model_output_d = convert_scmrun_to_plot_dict(model_output)

    tmp = []
    for group, run in target_d.items():
        model_unit = model_output_d[group].get_unique_meta("unit", True)
        run_converted = run.convert_unit(model_unit)

        tmp.append(run_converted)

    out = scmdata.run_append(tmp)

    return out



"""
Model runner class
"""
from __future__ import annotations

from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Protocol,
    Tuple,
    Union,
)

import numpy as np
import openscm_units
from attrs import define

if TYPE_CHECKING:
    import numpy.typing as nptype
    import pint
    import scmdata.run


class XToNamedPintConvertor(Protocol):  # pylint: disable=too-few-public-methods
    """
    Callable that supports converting the x-array to Pint quantities
    """

    def __call__(
        self,
        x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    ) -> Dict[str, Union[pint.Quantity[np.float_], np.float_]]:
        """
        Convert x to pint quantities
        """


class ModelRunsInputGenerator(Protocol):  # pylint: disable=too-few-public-methods
    """
    Callable that supports generating model run inputs
    """

    def __call__(
        self, **kwargs: Union[pint.Quantity[np.float_], np.float_]
    ) -> Dict[str, Any]:
        """
        Generate model run inputs
        """


class ModelRunner(Protocol):  # pylint: disable=too-few-public-methods
    """
    Callable that supports running the model
    """

    def __call__(self, **kwargs: Any) -> scmdata.run.BaseScmRun:
        """
        Run the model
        """


@define
class OptModelRunner:  # pylint: disable=too-few-public-methods
    """
    Model runner during optimisation
    """

    convert_x_to_names_with_units: XToNamedPintConvertor
    """
    Callable used to translate the x-array into input for ``do_model_runs_input_generator``

    This translates from the x-array used internally by e.g. scipy and
    emcee into a dictionary with meaningful keys and quantities with units (as
    needed). It must produce named output that can be passed directly to
    ``self.do_model_runs_input_generator``.
    """

    do_model_runs_input_generator: ModelRunsInputGenerator
    """
    Generator of inputs for ``do_model_runs``

    More specifically, the callable used to translate the parameters (already
    converted to :obj:`pint.Quantity`) into the keyword arguments required by
    ``self.do_model_runs``
    """

    do_model_runs: ModelRunner
    """
    Function that runs the model

    Runs the desired experiments based on inputs generated by
    ``self.do_model_runs_input_generator``
    """

    @classmethod
    def from_parameters(
        cls,
        params: Iterable[Tuple[str, Union[str, pint.Unit, None]]],
        do_model_runs_input_generator: ModelRunsInputGenerator,
        do_model_runs: ModelRunner,
    ) -> OptModelRunner:
        """
        Initialise from list of parameters

        This is a convenience method

        Parameters
        ----------
        params
            List of parameters

        do_model_runs_input_generator
            Generator of input for ``do_model_runs``. See docstring of
            ``self`` for more details

        do_model_runs
            Callable which does the model runs. See docstring of  ``self`` for
            more details

        Returns
        -------
            Initialised :obj:`OptModelRunner`
        """
        convert_x_to_names_with_units = partial(
            x_and_parameters_to_named_with_units,
            params=params,
        )

        return OptModelRunner(
            convert_x_to_names_with_units=convert_x_to_names_with_units,
            do_model_runs_input_generator=do_model_runs_input_generator,
            do_model_runs=do_model_runs,
        )

    def run_model(
        self,
        x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    ) -> scmdata.run.BaseScmRun:
        """
        Run the model

        Parameters
        ----------
        x
            Vector of calibration parameter values

        Returns
        -------
            Results of run
        """
        x_converted_name = self.convert_x_to_names_with_units(x)

        do_model_runs_inputs = self.do_model_runs_input_generator(**x_converted_name)

        res = self.do_model_runs(**do_model_runs_inputs)

        return res


def x_and_parameters_to_named_with_units(
    x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    params: Iterable[Tuple[str, Union[str, pint.Unit, None]]],
    get_unit_registry: Optional[Callable[[], pint.UnitRegistry]] = None,
) -> Dict[str, Union[pint.Quantity[np.float_], np.float_]]:
    """
    Convert x array and parameters to a dictionary and add units

    Parameters
    ----------
    x
        Vector of calibration parameter values

    params
        params to be calibrated

        This defines both the names and units of the parameters. If the units
        are ``None``, the values are assumed to be plain numpy quantities.

    get_unit_registry
        Function to get unit registry. This allows the user to do a delayed
        import of the unit registry which is important because pint's unit
        registries don't parallelise well.

        If not provided, ``openscm_units.unit_registry`` is used

    Returns
    -------
        Parameters, named and converted to :obj:`pint.Quantity` where
        appropriate

    Examples
    --------
    It also possible to inject a different registry as needed
    >>> import pint
    >>> ur_plus_pop = pint.UnitRegistry()
    >>> ur_plus_pop.define("thousands = [population]")
    >>> def get_ur_with_pop():
    ...     return ur_plus_pop
    >>> # Withoout the injection, an error is raised
    >>> x_and_parameters_to_named_with_units(
    ...     [1.1, 3.2],
    ...     [("para_a", "m"), ("pop_weight", "thousands")],
    ... )
    Traceback (most recent call last):
    ...
    pint.errors.UndefinedUnitError: 'thousands' is not defined in the unit registry
    >>> # With the injection, this works nicely
    >>> x_and_parameters_to_named_with_units(
    ...     [1.1, 3.2, 4.0],
    ...     [("para_a", "m"), ("pop_weight", "thousands"), ("factor", None)],
    ...     get_ur_with_pop,
    ... )
    {'para_a': <Quantity(1.1, 'meter')>, 'pop_weight': <Quantity(3.2, 'thousands')>, 'factor': 4.0}
    """
    unit_reg = get_unit_registry() if get_unit_registry else openscm_units.unit_registry

    out = {}
    for val, (key, unit) in zip(x, params):
        if unit is not None:
            val_out = unit_reg.Quantity(val, unit)
        else:
            val_out = val

        out[key] = val_out

    return out



"""
Support for plotting emcee during run and afterwards
"""
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from openscm_calibration.emcee_utils import get_labelled_chain_data

try:
    import corner

    HAS_CORNER = True
except ImportError:  # pragma: no cover
    HAS_CORNER = False

try:
    import seaborn as sns

    HAS_SEABORN = True
except ImportError:  # pragma: no cover
    HAS_SEABORN = False

if TYPE_CHECKING:
    import emcee
    import matplotlib
    import numpy.typing as nptype


def plot_chains(  # pylint:disable=too-many-arguments
    inp: Union[emcee.backends.Backend],
    burnin: int,
    parameter_order: List[str],
    neg_log_likelihood_name: str,
    axes_d: Dict[str, matplotlib.axes.Axes],
    get_neg_log_likelihood_ylim: Optional[
        Callable[[nptype.NDArray[Union[np.float_, np.int_]]], Tuple[float, float]]
    ] = None,
    **kwargs: Any,
) -> None:
    """
    Plot chains in MCMC run

    The y-limits of the log likelihood axis are set dynamically

    Parameters
    ----------
    inp
        Object from which to plot the state

    burnin
        Number of iterations to treat as burn in

    parameter_order
        Order of model parameters. This must match the order used by  ``inp``.

    neg_log_likelihood_name
        Label when plotting negative log likelihood. Must match the expected
        name in ``axes_d``

    axes_d
        Axes on which to plot the chains. Must have a key for each name in
        ``parameter_order`` plus a key equal to the value of
        ``neg_log_likelihood_name``

    get_neg_log_likelihood_ylim
        Function which gets the y-limits for the negative log likelihood plot
        based on the costs. If not provided,
        :func:`get_neg_log_likelihood_ylim_default` is used

    **kwargs
        Passed to :func:`plot_parameter_chains`
    """
    labelled_data = get_labelled_chain_data(
        inp,
        parameter_order,
        neg_log_likelihood_name,
        burnin=0,
        thin=1,
    )

    if get_neg_log_likelihood_ylim is None:
        get_neg_log_likelihood_ylim = get_neg_log_likelihood_ylim_default

    for label, to_plot in labelled_data.items():
        ax = axes_d[label]  # pylint:disable=invalid-name
        plot_parameter_chains(
            ax,
            to_plot,
            burnin=burnin,
            **kwargs,
        )
        ax.set_ylabel(label)

        if label == neg_log_likelihood_name:
            ax.set_ylim(*get_neg_log_likelihood_ylim(to_plot))


def plot_parameter_chains(  # pylint:disable=too-many-arguments
    ax: matplotlib.Axes.axes,  # pylint:disable=invalid-name
    chain_values: nptype.NDArray[Union[np.float_, np.int_]],
    burnin: int,
    alpha_chain: float = 0.3,
    linewidth: float = 0.5,
    color: str = "0.2",
    alpha_vspan: float = 0.3,
    kwargs_chain: Optional[Dict[str, Any]] = None,
    kwargs_vspan: Optional[Dict[str, Any]] = None,
) -> matplotlib.Axes.axes:
    """
    Plot chains for a single parameter in an MCMC run

    Parameters
    ----------
    ax
        Axes on which to plot

    chain_values
        Chain values to plot (should be 2D)

    burnin
        Number of iterations to treat as burn in

    alpha_chain
        Alpha to use for the chains

    linewidth
        Linewidth to use for the chains

    color
        Colour to use for the chains

    alpha_vspan
        Alpha to use for the vertical span (which shows the burnin period)

    kwargs_chain
        Pass to :meth:`ax.plot` when plotting the chains

    kwargs_vspan
        Pass to :meth:`ax.axvspan` when plotting the vertical span
    """
    if not kwargs_chain:
        kwargs_chain = {}

    if not kwargs_vspan:
        kwargs_vspan = {}

    ax.plot(
        chain_values,
        lw=linewidth,
        alpha=alpha_chain,
        color=color,
        **kwargs_chain,
    )
    # If burnin is 2, then we burn the first two steps
    # i.e. 0 and 1 so the span should cover 0 and 1, but
    # not 2 i.e. the span covers burnin -1, but not burnin.
    # To make this visually easier to see, we do a span that
    # ends halfway between the last step in burnin and the first
    # in the chain i.e. we end at burnin - 0.5
    ax.axvspan(0, burnin - 0.5, alpha=alpha_vspan, **kwargs_vspan)
    ax.set_xlim(0, chain_values.shape[0])

    return ax


def get_neg_log_likelihood_ylim_default(
    neg_ll_values: nptype.NDArray[Union[np.float_, np.int_]],
    median_scaling: float = 1.5,
    max_scaling: float = 2.0,
) -> Tuple[float, float]:
    r"""
    Get the y-limits for the negative log likelihood axes

    This is the default algorithm

    Parameters
    ----------
    neg_ll_values
        Negative log likelihood values being plotted

    median_scaling
        Scaling to apply to the median value

    max_scaling
        Scaling to apply to the maximum value

    Returns
    -------
        y-limits

    Notes
    -----
    The algorithm is

    .. math::

        \text{median_scaled} = \text{median_scaling} \times \text{med}(\text{neg_ll_values}) \\
        \text{max_scaled} = \text{max_scaling} \times \max(\text{neg_ll_values}) \\
        \text{ymin} = \min(0, \text{median_scaled}, \text{max_scaled}) \\
        \text{ymax} = \max(0, \text{median_scaled}, \text{max_scaled})
    """
    median_scaled = float(median_scaling * np.median(neg_ll_values))
    if not np.isfinite(median_scaled):
        median_scaled = 0.0

    max_scaled = float(max_scaling * np.max(neg_ll_values))
    if not np.isfinite(max_scaled):
        max_scaled = 0.0

    ymin = float(
        min(
            0.0,
            median_scaled,
            max_scaled,
        )
    )
    ymax = float(
        max(
            0.0,
            median_scaled,
            max_scaled,
        )
    )

    return (ymin, ymax)


def plot_dist(  # pylint:disable=too-many-arguments
    inp: Union[emcee.backends.Backend],
    burnin: int,
    thin: int,
    parameter_order: List[str],
    axes_d: Dict[str, matplotlib.axes.Axes],
    common_norm: bool = False,
    fill: bool = True,
    legend: bool = False,
    **kwargs: Any,
) -> None:
    """
    Plot distributions from MCMC run

    This is a thin wrapper around :func:`sns.kdeplot` that sets helpful
    defaults.

    Parameters
    ----------
    inp
        Object from which to plot the state

    burnin
        Number of iterations to treat as burn in

    thin
        Thinning to use when sampling the chains

    parameter_order
        Order of model parameters

    axes_d
        Axes on which to plot the distributions

    common_norm
        Should all the distributions use the same normalisation? We generally
        set this to ``False`` as we want to see the chains independently. See
        docstring of :func:`sns.kdeplot` for full details.

    fill
        Should the KDE plots be filled? See docstring of :func:`sns.kdeplot`
        for full details.

    legend
        Should a legend be added to the plots? The legend is pretty
        meaningless (it is just the chain numbers) so we generally set this to
        ``False``. See docstring of :func:`sns.kdeplot` for full details.

    **kwargs
        Passed to :func:`sns.kdeplot`.
    """
    if not HAS_SEABORN:
        raise ImportError("``plot_dist`` requires seaborn to be installed")

    burnt_in_samples_labelled = get_labelled_chain_data(
        inp,
        parameter_order,
        burnin=burnin,
        thin=thin,
    )

    for label, to_plot in burnt_in_samples_labelled.items():
        ax = axes_d[label]  # pylint:disable=invalid-name

        sns.kdeplot(
            data=to_plot,
            ax=ax,
            common_norm=common_norm,
            fill=fill,
            legend=legend,
            **kwargs,
        )
        ax.set_xlabel(label)


DEFAULT_PLOT_CORNER_TITLE_KWARGS: Dict[str, Any] = {"fontsize": 12}
"""
Default value for ``title_kwargs`` used by ``plot_corner``

Provided to give the user an easy to modify these defaults if they wish or a
starting point
"""


DEFAULT_PLOT_CORNER_LABEL_KWARGS: Dict[str, Any] = {"fontsize": "x-small"}
"""
Default value for ``label_kwargs`` used by ``plot_corner``

Provided to give the user an easy to modify these defaults if they wish or a
starting point
"""


def plot_corner(  # pylint:disable=too-many-arguments,too-many-locals
    inp: Union[emcee.backends.Backend],
    burnin: int,
    thin: int,
    parameter_order: List[str],
    fig: matplotlib.figure.Figure,
    bins: int = 30,
    plot_contours: bool = True,
    smooth: bool = True,
    quantiles: Sequence[float] = (0.05, 0.17, 0.5, 0.83, 0.95),
    show_titles: bool = True,
    title_quantiles: Sequence[float] = (0.05, 0.5, 0.95),
    title_kwargs: Optional[Dict[str, Any]] = None,
    title_fmt: str = ".3f",
    label_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> None:
    """
    Plot corner plot of distribution from MCMC run

    This is a thin wrapper around :func:`corner.corner` which uses some
    sensible defaults.

    Parameters
    ----------
    inp
        Object from which to plot the state

    burnin
        Number of iterations to treat as burn in

    thin
        Thinning to use when sampling the chains

    parameter_order
        Order of model parameters

    fig
        Figure to use for plotting, should be empty i.e. have been cleared
        before being passed here

    bins
        Number of bins to use in histograms. See docstring of
        :func:`corner.corner` for full details.

    plot_contours
        Whether to plot contours on the 2D distribution plots. See docstring
        of :func:`corner.corner` for full details.

    smooth
        Whether to smooth the contours on the 2D distribution plots. See
        docstring of :func:`corner.corner` for full details.

    quantiles
        Quantiles at which to draw vertical lines in the histogram plots. See
        docstring of :func:`corner.corner` for full details.

    show_titles
        Whether to show titles on the histogram plots. See docstring of
        :func:`corner.corner` for full details.

    title_quantiles
        Quantiles to put in the titles of the histogram plots. See docstring
        of :func:`corner.corner` for full details.

    title_kwargs
        Keyword arguments to use when making the titles on the histogram
        plots. If not supplied, we use
        ``DEFAULT_PLOT_CORNER_TITLE_KWARGS``.

    title_fmt
        Format string to use when creating the titles on the histogram plots.
        If not supplied, our own internal defaults are used (see source code
        for values). See docstring of :func:`corner.corner` for full details.

    label_kwargs
        Keyword arguments to use when creating the labels for the plot. If not
        supplied, we use ``DEFAULT_PLOT_CORNER_LABEL_KWARGS``.

    **kwargs
        Passed to :func:`corner.corner`
    """
    if not HAS_CORNER:
        raise ImportError("``plot_corner`` requires corner to be installed")

    if title_kwargs is None:
        title_kwargs = DEFAULT_PLOT_CORNER_TITLE_KWARGS

    if label_kwargs is None:
        label_kwargs = DEFAULT_PLOT_CORNER_LABEL_KWARGS

    burnt_in_samples_flat = inp.get_chain(discard=burnin, thin=thin, flat=True)

    corner.corner(
        burnt_in_samples_flat,
        labels=parameter_order,
        fig=fig,
        bins=bins,
        plot_contours=plot_contours,
        smooth=smooth,
        quantiles=quantiles,
        show_titles=show_titles,
        title_quantiles=title_quantiles,
        title_kwargs=title_kwargs,
        title_fmt=title_fmt,
        label_kwargs=label_kwargs,
        **kwargs,
    )



"""
Storage class
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, MutableSequence, Protocol, Tuple, Union

import numpy as np
from attrs import define, field

if TYPE_CHECKING:
    import attr
    import numpy.typing as nptype
    import scmdata.run


class SupportsListLikeHandling(Protocol):  # pylint: disable=too-few-public-methods
    """
    Class that supports handling a list-like
    """

    def list(self, list_to_handle: MutableSequence[Any]) -> MutableSequence[Any]:
        """
        Get a new object that behaves like a :obj:`MutableSequence`
        """


def _all_none_to_start(
    instance: OptResStore,  # pylint: disable=unused-argument
    attribute: attr.Attribute[MutableSequence[Any]],
    value: MutableSequence[Any],
) -> None:
    """
    Check all values are ``None``

    Parameters
    ----------
    self
        Object instance

    attribute
        Attribute to check

    value
        Value to check

    Raises
    ------
    ValueError
        Not all elements in ``value`` are ``None``
    """
    if not all(v is None for v in value):
        raise ValueError(
            f"All values in ``{attribute.name}`` should be ``None`` to start"
        )


def _same_length_as_res(
    instance: OptResStore,
    attribute: attr.Attribute[MutableSequence[Any]],
    value: MutableSequence[Any],
) -> None:
    """
    Check ``value`` has same length as ``instance.res``

    Parameters
    ----------
    self
        Object instance

    attribute
        Attribute to check

    value
        Value to check

    Raises
    ------
    ValueError
        ``value`` does not have the same length as ``instance.res``
    """
    if len(value) != len(instance.res):
        raise ValueError(f"``{attribute.name}`` must be the same length as ``res``")


def _contains_indices_in_res(
    instance: OptResStore,
    attribute: attr.Attribute[MutableSequence[int]],
    value: MutableSequence[int],
) -> None:
    """
    Check ``value`` has indices that line up with ``instance.res``

    Parameters
    ----------
    self
        Object instance

    attribute
        Attribute to check

    value
        Value to check

    Raises
    ------
    ValueError
        ``value`` does not have indices that line up with ``instance.res``
    """
    exp_indices = list(range(len(instance.res)))
    if list(sorted(value)) != exp_indices:
        raise ValueError(
            f"{attribute.name} must have indices: {exp_indices}, received: {value}"
        )


@define
class OptResStore:
    """
    Store for results during optimisation
    """

    res: MutableSequence[Union[None, scmdata.run.BaseScmRun]] = field(
        validator=[_all_none_to_start]
    )
    """Results of runs"""

    costs: MutableSequence[Union[None, float]] = field(
        validator=[_all_none_to_start, _same_length_as_res]
    )
    """Costs of runs"""

    x_samples: MutableSequence[
        Union[None, nptype.NDArray[Union[np.float_, np.int_]]]
    ] = field(validator=[_all_none_to_start, _same_length_as_res])
    """x vectors sampled"""

    params: Tuple[str]
    """Names of the parameters being stored in ``x_samples``"""

    available_indices: MutableSequence[int] = field(
        validator=[_same_length_as_res, _contains_indices_in_res]
    )
    """Indices available to be written into"""

    @classmethod
    def from_n_runs(
        cls,
        n_runs: int,
        params: Tuple[str],
    ) -> OptResStore:
        """
        Initialise based on expected number of runs

        Parameters
        ----------
        n_runs
            Expected number of runs

        params
            Names of the parameters that are being sampled

        Returns
        -------
            Initialised store
        """
        # Reverse so that using pop counts up
        available_indices = list(range(n_runs))[::-1]
        return cls(
            res=[None] * n_runs,
            costs=[None] * n_runs,
            x_samples=[None] * n_runs,
            params=params,
            available_indices=available_indices,
        )

    @classmethod
    def from_n_runs_manager(
        cls,
        n_runs: int,
        manager: SupportsListLikeHandling,
        params: Tuple[str],
    ) -> OptResStore:
        """
        Initialise based on expected number of runs for use in parallel work

        Parameters
        ----------
        n_runs
            Expected number of runs

        manager
            Manager of lists (e.g. :class:`multiprocess.managers.SyncManager`)

        params
            Names of the parameters that are being sampled

        Returns
        -------
            Initialised store
        """
        # Reverse so that using pop counts up
        available_indices = list(range(n_runs))[::-1]

        return cls(
            res=manager.list([None] * n_runs),
            costs=manager.list([None] * n_runs),
            x_samples=manager.list([None] * n_runs),
            params=params,
            available_indices=manager.list(available_indices),
        )

    def get_available_index(self) -> int:
        """
        Get an available index to write into

        Returns
        -------
            Available index. This index is now no longer considered available.
        """
        return self.available_indices.pop()  # pylint:disable=no-member

    def set_result_cost_x(
        self,
        res: Union[None, scmdata.run.BaseScmRun],
        cost: float,
        x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
        idx: int,
    ) -> None:
        """
        Set result, cost and x at a given index

        Parameters
        ----------
        res
            Result to append (use ``None`` for a failed run)

        cost
            Cost associated with the run

        x
            Parameter array associated with the run

        idx
            Index in ``self.costs``, ``self.x_samples`` and ``self.res`` to write into
        """
        if len(x) != len(self.params):
            raise ValueError(
                f"Length of parameter vector ({len(x)}) does not match "
                f"length of ``self.params`` ({len(self.params)})"
            )

        self.costs[idx] = cost  # pylint: disable=unsupported-assignment-operation
        self.x_samples[idx] = x  # pylint: disable=unsupported-assignment-operation
        self.res[idx] = res  # pylint: disable=unsupported-assignment-operation

    def append_result_cost_x(
        self,
        res: scmdata.run.BaseScmRun,
        cost: float,
        x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    ) -> None:
        """
        Append result, cost and x from a successful run to the results

        Parameters
        ----------
        res
            Result to append (use ``None`` for a failed run)

        cost
            Cost associated with the run

        x
            Parameter array associated with the run
        """
        iteration = self.get_available_index()
        res_keep = res.copy()
        res_keep["it"] = iteration

        self.set_result_cost_x(
            res=res_keep,
            cost=cost,
            x=x,
            idx=iteration,
        )

    def note_failed_run(
        self,
        cost: float,
        x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    ) -> None:
        """
        Note that a run failed

        Typically, ``cost`` will be ``np.inf``.

        The cost and x parameters are appended to the results, as well as an
        indicator that the run was a failure in ``self.res``.

        Parameters
        ----------
        cost
            Cost associated with the run

        x
            Parameter array associated with the run
        """
        iteration = self.get_available_index()
        self.set_result_cost_x(
            res=None,
            cost=cost,
            x=x,
            idx=iteration,
        )

    def get_costs_xsamples_res(
        self,
    ) -> Tuple[
        Tuple[float, ...],
        Tuple[nptype.NDArray[Union[np.float_, np.int_]], ...],
        Tuple[scmdata.run.BaseScmRun, ...],
    ]:
        """
        Get costs, x_samples and res from runs

        Returns
        -------
            Costs, x_samples and res from all runs which were attempted (i.e. we
            include failed runs here)
        """
        # There may be a better algorithm for this, PRs welcome :)
        if all(x is None for x in self.x_samples):  # pylint: disable=not-an-iterable
            return ((), (), ())

        tmp = tuple(
            zip(
                *[
                    (
                        self.costs[i],  # pylint: disable=unsubscriptable-object
                        x,
                        self.res[i],  # pylint: disable=unsubscriptable-object
                    )
                    for i, x in enumerate(self.x_samples)
                    # x is only None if no run was attempted yet
                    if x is not None
                ]
            )
        )

        # Help out type hinting
        costs: Tuple[float, ...] = tmp[0]
        xs_out: Tuple[nptype.NDArray[Union[np.float_, np.int_]], ...] = tmp[1]
        ress: Tuple[scmdata.run.BaseScmRun, ...] = tmp[2]

        out = (costs, xs_out, ress)

        return out

    def get_costs_labelled_xsamples_res(
        self,
    ) -> Tuple[
        Tuple[float, ...],
        Dict[str, nptype.NDArray[Union[np.float_, np.int_]]],
        Tuple[scmdata.run.BaseScmRun, ...],
    ]:
        """
        Get costs, x_samples and res from runs

        Returns
        -------
            Costs, x_samples and res from all runs which were attempted (i.e. we
            include failed runs here)
        """
        unlabelled = self.get_costs_xsamples_res()
        if not any(unlabelled):
            return (
                unlabelled[0],
                {p: np.array([]) for p in self.params},
                unlabelled[2],
            )

        x_samples_stacked = np.vstack(unlabelled[1])
        xs_labelled = {p: x_samples_stacked[:, i] for i, p in enumerate(self.params)}

        out = (unlabelled[0], xs_labelled, unlabelled[2])

        return out



"""
Helpers for emcee
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Union

    # See here for explanation of this pattern and why we don't need quotes
    # below https://docs.python.org/3/library/typing.html#constant
    import emcee.backends
    import numpy.typing as nptype


def get_acceptance_fractions(
    chains: nptype.NDArray[np.float_],
) -> nptype.NDArray[np.float_]:
    """
    Get acceptance fraction in each chain of an MCMC ensemble of chains

    Parameters
    ----------
    chains
        Chains. We expected that the the axes are ["step", "chain",
        "parameter"]

    Returns
    -------
        Acceptance fraction in each chain
    """
    # This is complicated, in short:
    # 1. find differences between steps across all calibration parameters
    # 2. check, in each chain, if there is any difference in any of the
    #   parameter values (where this happens, it means that the step was
    #   accepted)
    # 3. sum up the number of accepted steps in each chain
    accepted: nptype.NDArray[np.int_] = np.sum(
        np.any(np.diff(chains, axis=0), axis=2), axis=0
    )
    n_proposals = chains.shape[0] - 1  # first step isn't a proposal
    acceptance_fractions = accepted / np.float_(n_proposals)

    return acceptance_fractions


def get_autocorrelation_info(
    inp: emcee.backends.Backend,
    burnin: int,
    thin: int = 1,
    autocorr_tol: int = 0,
    convergence_ratio: float = 50,
) -> Dict[str, Union[float, int, bool, nptype.NDArray[np.float_]]]:
    """
    Get info about autocorrelation in chains

    Parameters
    ----------
    inp
        Object of which to check autocorrelation

    burnin
        Number of iterations to treat as burn-in

    thin
        Thinning to apply to the chains. Emcee handles
        this so that the returned output is in steps,
        not thinned steps.

    autocorr_tol
        Tolerance for auto-correlation calculations. Set
        to zero to force calculation to never fail

    convergence_ratio
        If the number of iterations (excluding burn-in) is
        greater than ``convergence_ratio`` multiplied by
        the autocorrelation (averaged over all the chains),
        we assume the chains have converged

    Returns
    -------
        Results of calculation, keys:

        - "tau": autocorrelation in each chains
        - "autocorr": average of tau
        - "converged": whether the chains have converged or not based on
          ``convergence_ratio``
        - "convergence_ratio": value of ``convergence_ratio``
        - "steps_post_burnin": Number of steps in chains post burn-in

    """
    tau = inp.get_autocorr_time(discard=burnin, tol=autocorr_tol, thin=thin)
    autocorr = np.mean(tau)

    converged = inp.iteration - burnin > convergence_ratio * autocorr

    out = {
        "tau": tau,
        "autocorr": autocorr,
        "converged": converged,
        "convergence_ratio": convergence_ratio,
        "steps_post_burnin": inp.iteration - burnin,
    }

    return out


def get_labelled_chain_data(
    inp: Union[emcee.backends.Backend],
    parameter_order: List[str],
    neg_log_likelihood_name: Optional[str] = None,
    burnin: int = 0,
    thin: int = 0,
) -> Dict[str, nptype.NDArray[Union[np.float_, np.int_]]]:
    """
    Get labelled chain data

    Parameters
    ----------
    inp
        Object from which to plot the state

    parameter_order
        Order of model parameters. This must match the order used by  ``inp``.

    neg_log_likelihood_name
        Name to use for the negative log likelihood data. If not provided,
        negative log likelihood information is not returned.

    burnin
        Number of iterations to treat as burn in

    thin
        Thinning to use when sampling the chains
    """
    all_samples = inp.get_chain(discard=burnin, thin=thin)

    out = {para: all_samples[:, :, i] for i, para in enumerate(parameter_order)}

    if neg_log_likelihood_name:
        all_neg_ll = inp.get_log_prob(discard=burnin, thin=thin)
        out[neg_log_likelihood_name] = all_neg_ll

    return out



"""
Minimisation helpers
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Protocol, Type, Union

import numpy as np
import scmdata.run

from .store import OptResStore

if TYPE_CHECKING:
    import numpy.typing as nptype


class SupportsCostCalculation(Protocol):  # pylint: disable=too-few-public-methods
    """
    Class that supports cost calculations
    """

    def calculate_cost(self, model_results: scmdata.run.BaseScmRun) -> float:
        """
        Calculate cost function

        Parameters
        ----------
        model_results
            Model results for which to calculate the cost

        Returns
        -------
            Cost function value
        """


class SupportsModelRun(Protocol):  # pylint: disable=too-few-public-methods
    """
    Class that supports model runs
    """

    def run_model(
        self,
        x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    ) -> scmdata.run.BaseScmRun:
        """
        Calculate cost function

        Parameters
        ----------
        x
            Parameter values

        Returns
        -------
            results of model run
        """


def to_minimize_full(
    x: nptype.NDArray[Union[np.float_, np.int_]],  # pylint: disable=invalid-name
    cost_calculator: SupportsCostCalculation,
    model_runner: SupportsModelRun,
    store: Optional[OptResStore] = None,
    known_error: Optional[Type[ValueError]] = None,
) -> float:
    """
    Calculate cost for given set of parameters

    This is a function that can be minimised with scipy

    Before passing to scipy, all the arguments except ``x`` should be filled
    using e.g. :func:`functools.partial`

    Parameters
    ----------
    x
        Parameter array

    cost_calculator
        Calculator of the cost function for each set of model results

    model_runner
        Model runner

    store
        Store of results at each step in the optimisation (useful for plotting)

    known_error
        Known error that can occur when solving. If any other error is
        encountered, it is raised rather than being allowed to pass.

    Returns
    -------
        Cost function for array ``x``
    """
    if known_error:
        try:
            model_results = model_runner.run_model(x)
        except known_error:
            cost = np.inf
            if store:
                store.note_failed_run(cost, x)
            return cost
    else:
        model_results = model_runner.run_model(x)

    cost = cost_calculator.calculate_cost(model_results)
    if store:
        store.append_result_cost_x(model_results, cost, x)

    return cost



"""
Iterable utility functions
"""
from __future__ import annotations

import itertools
from typing import Any, Iterable, List


def repeat_elements(inp: Iterable[Any], n_rep: int) -> List[Any]:
    """
    Repeat elements of an iterable ``n`` times

    Parameters
    ----------
    inp
        Iterable of which to repeat the elements

    n_rep
        Number of times to repeat

    Returns
    -------
        List with repeated elements

    Examples
    --------
    >>> repeat_elements(["a", "b"], 2)
    ['a', 'a', 'b', 'b']
    """
    return list(
        itertools.chain(*[list(itertools.repeat(element, n_rep)) for element in inp])
    )



"""
Calculate cost of model results
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from attrs import define, field

if TYPE_CHECKING:
    import attr
    import pandas as pd
    import scmdata.run


def _works_with_self_target(
    instance: OptCostCalculatorSSE,
    attribute: attr.Attribute[scmdata.run.BaseScmRun],
    value: scmdata.run.BaseScmRun,
) -> None:
    def _get_msg() -> str:
        target_ts = instance.target.timeseries()
        value_ts = value.timeseries()

        msg = (
            "target and normalisation are somehow misaligned "
            "(passing self.target to self.calculate_cost results "
            "in nan), please check.\n"
            f"target timeseries:\n{target_ts}\n"
            f"{attribute.name} timeseries:\n{value_ts}"
        )

        return msg

    try:
        instance.calculate_cost(instance.target)
    except KeyError as exc:
        raise ValueError(_get_msg()) from exc


def _is_meta_in_target(
    instance: OptCostCalculatorSSE,
    attribute: attr.Attribute[str],
    value: str,
) -> None:
    available_metadata = instance.target.meta_attributes
    if value not in available_metadata:
        msg = (
            f"value of ``{attribute.name}``, '{value}' is not in the metadata "
            f"of target. Available metadata: {available_metadata}"
        )

        raise KeyError(msg)


@define
class OptCostCalculatorSSE:  # pylint: disable=too-few-public-methods
    """
    Cost calculator based on sum of squared errors

    This is a convenience class. We may want to refactor it in future to
    provide greater flexibility for other cost calculations.
    """

    target: scmdata.run.BaseScmRun
    """Target timeseries"""

    model_col: str = field(validator=[_is_meta_in_target])
    """
    Column which contains the name of the model.

    This is used when subtracting the model results from the target
    """

    normalisation: scmdata.run.BaseScmRun = field(validator=[_works_with_self_target])
    """
    Normalisation values

    Should have same timeseries as target. See the class methods for helpers.
    """

    @classmethod
    def from_unit_normalisation(
        cls, target: scmdata.run.BaseScmRun, model_col: str
    ) -> OptCostCalculatorSSE:
        """
        Initialise assuming unit normalisation for each timeseries.

        This is a convenience method, but is not recommended for any serious
        work as unit normalisation is unlikely to be a good choice for most
        problems.

        Parameters
        ----------
        target
            Target timeseries

        model_col
            Column which contains of the model in ``target``

        Returns
        -------
            :obj:`OptCostCalculatorSSE` such that the normalisation is 1 for
            all timepoints (with the units defined by whatever the units of
            each timeseries are in ``target``)
        """
        norm = target.timeseries()
        norm.loc[:, :] = 1
        norm = type(target)(norm)

        return cls(target=target, normalisation=norm, model_col=model_col)

    @classmethod
    def from_series_normalisation(
        cls,
        target: scmdata.run.BaseScmRun,
        model_col: str,
        normalisation_series: pd.Series,
    ) -> OptCostCalculatorSSE:
        """
        Initialise starting from a series that defines normalisation for each timeseries.

        The series is broadcast to match the timeseries in target, using the
        same value for all timepoints in each timeseries.

        Parameters
        ----------
        target
            Target timeseries

        model_col
            Column which contains of the model in ``target``

        normalisation_series
            Series to broadcast to create the desired normalisation

        Returns
        -------
            Initialised :obj:`OptCostCalculatorSSE`
        """
        required_columns = {"variable", "unit"}
        missing_cols = required_columns - set(normalisation_series.index.names)
        if missing_cols:
            msg = (
                "normalisation is missing required column(s): "
                f"``{sorted(missing_cols)}``"
            )
            raise KeyError(msg)

        target_ts_no_unit = target.timeseries().reset_index("unit", drop=True)

        # This is basically what pandas does internally when doing ops:
        # align and then broadcast
        norm_series_aligned, _ = normalisation_series.align(target_ts_no_unit)

        if norm_series_aligned.isnull().any().any():
            msg = (
                "Even after aligning, there are still nan values.\n"
                f"norm_series_aligned:\n{norm_series_aligned}\n"
                f"target_ts_no_unit:\n{target_ts_no_unit}"
            )
            raise ValueError(msg)

        if norm_series_aligned.shape[0] != target_ts_no_unit.shape[0]:
            msg = (
                "After aligning, there are more rows in the normalisation "
                "than in the target.\n"
                f"norm_series_aligned:\n{norm_series_aligned}\n"
                f"target_ts_no_unit:\n{target_ts_no_unit}"
            )
            raise ValueError(msg)

        norm_series_aligned = type(target_ts_no_unit)(
            np.broadcast_to(norm_series_aligned.values, target_ts_no_unit.T.shape).T,
            index=norm_series_aligned.index,
            columns=target_ts_no_unit.columns,
        )

        normalisation = type(target)(norm_series_aligned)

        return cls(target=target, normalisation=normalisation, model_col=model_col)

    def calculate_cost(self, model_results: scmdata.run.BaseScmRun) -> float:
        """
        Calculate cost function based on model results

        Parameters
        ----------
        model_results
            Model results of which to calculate the cost

        Returns
        -------
            Cost
        """
        diff = model_results.subtract(
            self.target, op_cols={self.model_col: "res - target"}
        ).divide(
            self.normalisation,
            op_cols={self.model_col: "(res - target) / normalisation"},
        )

        cost = float((diff.convert_unit("1") ** 2).values.sum().sum())

        return cost
